/*defines that control compilation*/
#define TI89_VERSION 0

#if TI89_VERSION == 1
#define USE_TI89
#define calc(x, y) (x)
#else
#define USE_TI92PLUS
#define USE_V200
#define calc(x, y) (y)
#endif

#include <tigcclib.h>

#define RANDOM_STRENGTH 4
#define VTILOG_DISABLED

#include "extrand.h"
#include "extgraph.h"
#include "vtilog.h"

/*defines that set constants in the code*/
#define TRIES_PER_COMP 32
#define QUEUE_SIZE 32
#define V_MAJOR 1
#define V_MINOR 2
#define DATA_NAME $(nyetwdat)

typedef struct
{
	char initials[3];
	unsigned char random_strength : 3, calc : 1, padding : 4;
} FILE_VERSION;

typedef struct
{
	unsigned char cursor : 1, server : 1, computer : 1, connected : 1, con_n : 1, con_e :
	1, con_s : 1, con_w : 1;
} SQUARE;

typedef struct
{
	char x, y;
} LOCATION;

typedef struct
{
	struct
	{
		short size;
		FILE_VERSION version;
	} header;
	struct
	{
		char height;
		char width;
		char save;
		short computer_moves;
		short difficulty;
		short score;
		short scores[21];
		RANDOM_BUFFER buffer;
		SQUARE level[calc(11, 14)][calc(19, 29)];
	} data;
	struct
	{
		char hash[17];
		char type[7];
	} footer;
} DATA_FILE;

static char MainMenu(short (*)[], short *);
static void MenuText(short (*)[], unsigned char[], short);
static short Wrap(short, short, short);
static short PlayGame(short, char *, char *, char *, short *, short *, short *, SQUARE[][calc(19, 29)]);
static void GenerateNetwork(char, char, char, char, SQUARE[][calc(19, 29)]);
static char SquareIsEmpty(SQUARE);
static void PlaceComputer(short, short, char, char, char, SQUARE[][calc(19, 29)]);
static void UpdateConnectable(char, char, SQUARE[][calc(19, 29)], char[][calc(19, 29)]);
static char Connections(SQUARE);
static char FindValidDirections(char[], short, short, char, char, char, char[][calc(19, 29)]);
static char Connect(short *, short *, char, char, char, SQUARE[][calc(19, 29)]);
static void RefreshDisplay(char, char, SQUARE[][calc(19, 29)]);
static void DrawSprite(char, char, char, char, SQUARE);
static void UpdateConnected(char, char, SQUARE[][calc(19, 29)]);
static char CheckForWin(char, char, SQUARE[][calc(19, 29)]);
static short ScrambleLevel(char, char, SQUARE[][calc(19, 29)]);
static void RotateSquare(SQUARE *, char);
static void UnscrambleLevel(short *, short *, short *, char, char, SQUARE[][calc(19, 29)]);
static void CheckHighScores(short, short, short (*)[]);
static void Pause(void);

void _main(void)
{
	char hash[16];
	char escape = 0; /*used to break out of the main loop*/
	short old_font; /*used for stroing a backup of the font*/
	short score;
	DATA_FILE *filep = NULL; /*a pointer that is used in file access*/
	static FILE_VERSION Check = {{'P', 'J', 'R'}, RANDOM_STRENGTH, TI89_VERSION, 0}; /*a piece of data
	that is used to make sure the file belongs to this program*/
	INT_HANDLER OldInt1; /*used to turn off int1 so _rowread works better*/
	INT_HANDLER OldInt5; /*used to turn off int5 so _rowread works better*/
	LCD_BUFFER buffer;
	MD5_CTX ctx;
	SYM_ENTRY *file_vat = NULL; /*another pointer that is used for file creation and access*/

	FastCopyScreen_R(LCD_MEM, buffer);

	file_vat = SymFindPtr(DATA_NAME, 0); /*try to find the data file*/

	if(file_vat != NULL) /*if the file exists*/
	{
		if(file_vat->flags.bits.archived) /*if the file is archived*/
		{
			EM_moveSymFromExtMem(DATA_NAME, HS_NULL); /*unarchive the file*/
		}
		filep = HeapDeref(file_vat->handle); /*get a pointer to the beginning of
		the file*/
		MD5Init(&ctx);
		MD5Update(&ctx, (char*)&filep->data, sizeof(filep->data));
		MD5Final(hash, &ctx);
	}

	if(file_vat == NULL || memcmp(&filep->header.version, &Check, sizeof(FILE_VERSION))
	|| memcmp(hash, filep->footer.hash,	16))
	{
		RANDOM_BUFFER buffer; /*a variable used for creating a new file*/

		randomize_long(); /*scramble the random number generator*/

		SaveRandState(&buffer); /*save the state of the random number generator*/

		file_vat = DerefSym(SymAdd(DATA_NAME)); /*create the entry for the data file*/
		file_vat->handle = HeapAlloc(sizeof(DATA_FILE)); /*allocate the memory for the data file*/
		filep = HeapDeref(file_vat->handle); /*get a pointer to the beginning of the file*/
		memset(filep, 0, sizeof(DATA_FILE));

		filep->header.size = sizeof(DATA_FILE) - 2; /*put in the size bytes*/
		memcpy(&filep->header.version, &Check, sizeof(FILE_VERSION)); /*put in the
		version check*/

		memcpy(&filep->data.buffer, &buffer, sizeof(RANDOM_BUFFER));

		memcpy(filep->footer.type, &(unsigned char[]){0, 'D', 'a', 't', 'a', 0, OTH_TAG},
		7); /*put in the file type*/
	}

	LoadRandState(&filep->data.buffer); /*restore the state of the random number generator
	that was saved at exit*/

	old_font = FontSetSys(F_4x6); /*set the font to 4x6 and save the old font*/

	/*save and turn off auto ints 1 and 5*/
	OldInt1 = GetIntVec(AUTO_INT(1));
	SetIntVec(AUTO_INT(1), DUMMY_HANDLER);
	OldInt5 = GetIntVec(AUTO_INT(5));
	SetIntVec(AUTO_INT(5), DUMMY_HANDLER);

	do
	{
		if(filep->data.save == 0) /*if a save game does not exist*/
		{
			escape = MainMenu(&filep->data.scores, &filep->data.difficulty); /* go to the
			menu*/
		}
		if(escape != 1) /*if esc was not pressed*/
		{
			PlayGame(filep->data.difficulty, &filep->data.save, &filep->data.height, &filep->data.width,
			&score, &filep->data.score, &filep->data.computer_moves, filep->data.level); /*play the game*/
			if(score >= 0)
			{
				CheckHighScores(filep->data.difficulty, score, &filep->data.scores); /*see if you got a
				high score*/
			}
			else if(score == -2)
			{
				filep->data.save = 1;
			}
		}
	} while(escape != 1 && score != -2); /*while esc or clear was not pressed*/

	/*restore auto ints 1 and 5*/
	SetIntVec(AUTO_INT(1), OldInt1);
	SetIntVec(AUTO_INT(5), OldInt5);
	/*restore the font*/
	FontSetSys(old_font);
	/*save the state of the random number generator*/
	SaveRandState(&filep->data.buffer);

	/*produce a hash of the data section and write it to the file*/
	MD5Init(&ctx);
	MD5Update(&ctx, (char*)&filep->data, sizeof(filep->data));
	MD5Final(filep->footer.hash, &ctx);

	EM_moveSymToExtMem(DATA_NAME, HS_NULL); /*archive the data file*/

	FastCopyScreen_R(buffer, LCD_MEM);
	ST_helpMsg(EXTGRAPH_VERSION_PWDSTR);
}

char MainMenu(short (*HighScores)[3], short *difficulty)
{
	static unsigned char dot[3] = {0x70, 0x70, 0x70}; /*a sprite*/
	static unsigned char dot_x[3] = {10, calc(62, 102),	calc(139, 219)}; /*alignment
	data*/
	short key; /*used to store rowread values*/

	ClrScr();
	MenuText(HighScores, dot, *difficulty); /*draw all the text for the menu, including
	high scores*/

	for(;;)
	{
		key = _rowread(0xFFFE); /*get the arrow key row*/

		if(key & calc(10, 80)) /*if left or right was pressed*/
		{
			Sprite8_XOR(dot_x[*difficulty], 31, 3, dot, LCD_MEM); /*turn off the dot*/
		}
		if(key & calc(2, 16)) /*if left pressed*/
		{
			(*difficulty)--;
			while(_rowread(0xFFFE) & calc(2, 16)); /*wait until left not pressed*/
		}
		if(key & calc(8, 64)) /*if right pressed*/
		{
			(*difficulty)++;
			while(_rowread(0xFFFE) & calc(8, 64)); /*wait until right not pressed*/
		}
		if(key & calc(10, 80)) /*if left or right pressed*/
		{
			*difficulty = Wrap(*difficulty, 0, 2); /*put the difficulty in the correct
			range*/
			Sprite8_OR(dot_x[*difficulty], 31, 3, dot, LCD_MEM); /*put the dot back*/
		}
		if(key & calc(16, 1)) /*if 2nd pressed*/
		{
			while(_rowread(0xFFFE) & 16); /*wait until 2nd not pressed*/
			return(0);
		}
		if(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64)) /*if esc pressed*/
		{
			while(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64));
			return(1); /*tell _main to quit*/
		}
	}
}

void MenuText(short (*HighScores)[3], unsigned char dot[], short difficulty)
{
	unsigned char buffer[14]; /*a buffer for sprintf*/
	short i, j; /*loop counters*/
	static unsigned char circle[5] = {0x70, 0x88, 0x88, 0x88, 0x70}; /*a sprite*/
	static const char *ordinal[7] = {"1st", "2nd", "3rd", "4th", "5th", "6th", "7th"}; /*a
	list of the ordinal numbers*/
	static unsigned char offset[7] = {2, 0, 0, 1, 1, 1, 1}; /*alignment data for the place
	names*/
	static unsigned char dot_offset[3] = {10, calc(62, 102), calc(139, 219)}; /*alignment
	data for the dot*/

	sprintf(buffer, "Nyetwork v%d.%d", V_MAJOR, V_MINOR);
	DrawStr((calc(160, 240) - DrawStrWidth(buffer, F_4x6)) / 2, 0, buffer, A_NORMAL);
	DrawStr(calc(51, 91), 6, "By Peter J. Rowe", A_NORMAL);
	DrawStr(calc(49, 89), 12, "mig53@yahoo.com", A_NORMAL);
	DrawStr(calc(64, 104), 24, "Difficulty", A_NORMAL);
	DrawStr(15, 30, "Easy", A_NORMAL);
	DrawStr(calc(67, 107), 30, "Medium", A_NORMAL);
	DrawStr(calc(144, 224), 30, "Hard", A_NORMAL);
	DrawStr(calc(60, 100), 42, "High Scores", A_NORMAL);

	for(i = 0; i < 3; i++)
	{
		for(j = 0; j < 7; j++)
		{
			DrawStr(offset[j], 48 + 6 * j, ordinal[j], A_NORMAL);
			sprintf(buffer, "%d", HighScores[j][i]);
			DrawStr((i == 0 ? 15 : (i == 1 ? (calc(160, 240) - DrawStrWidth(buffer,
			F_4x6)) / 2 : calc(160, 240) - DrawStrWidth(buffer, F_4x6))), 48 + 6 * j,
			buffer, A_NORMAL);
		}
	}

	Sprite8_OR_R(10, 30, 5, circle, LCD_MEM);
	Sprite8_OR_R(dot_offset[difficulty], 31, 3, dot, LCD_MEM);
	Sprite8_OR_R(calc(62, 102), 30, 5, circle, LCD_MEM);
	Sprite8_OR_R(calc(139, 219), 30, 5, circle, LCD_MEM);
	DrawStr(calc(38, 78), 95, "Use \x15/\x16 + [2nd] or [ESC]", A_NORMAL);
}

short Wrap(short input, short lower, short upper) /*take the input value and put it
between the lower and upper bound with wrapping around the ends*/
{
	return(input >= lower ? (input <= upper ? input : lower) : upper);
}

short PlayGame(short difficulty, char *savep, char *heightp, char *widthp, short *scorep,
short *file_scorep, short *computer_movesp, SQUARE level[][calc(19, 29)])
{
	char computers = 0;
	char wrap = 0;

	ClrScr();

	if(*savep == 0)
	{
		DrawStr(calc(60, 100), calc(39, 53), "Please Wait", A_NORMAL);
		DrawStr(calc(49, 89), calc(45, 59), "Generating Level", A_NORMAL);

		switch(difficulty)
		{
			case 0:
			computers = calc(9 + random_long(10), 17 + random_long(18));
			*heightp = calc(6, 8);
			*widthp = calc(11, 16);
			wrap = 0;
			break;

			case 1:
			computers = calc(35 + random_long(22), 68 + random_long(41));
			*heightp = calc(11, 14);
			*widthp = calc(19, 29);
			wrap = 0;
			break;

			case 2:
			computers = calc(35 + random_long(22), 68 + random_long(41));
			*heightp = calc(11, 14);
			*widthp = calc(19, 29);
			wrap = 1;
		}

		memset(level, 0, calc(11, 14) * calc(19, 29) * sizeof(SQUARE));

		GenerateNetwork(computers, *heightp, *widthp, wrap, level);

		*scorep = (*computer_movesp = ScrambleLevel(*heightp, *widthp, level)) + (difficulty == 2 ? 20 :
		10) * computers + random_long(2); /*the random_long(2) makes it so that some of the scores are
		odd. without it, all the scores are even*/

		ClrScr();
	}
	else
	{
		*scorep = *file_scorep;
		*savep = 0;
	}

	FastOutlineRect_R(LCD_MEM, calc(78, 118) - 4 * (*widthp), calc(52, 68) - 4 * (*heightp), calc(81, 121)
	+ 4 * (*widthp), calc(55, 71) + 4 * (*heightp), A_NORMAL);

	UnscrambleLevel(scorep, file_scorep, computer_movesp, *heightp, *widthp, level); /*have the player do
	thier thing*/

	return(*scorep);
}

void GenerateNetwork(char computers, char height, char width, char wrap, SQUARE
level[][calc(19, 29)])
{
	char buffer[10]; /*a buffer for sprintf*/
	short x, y; /*these are temporary variables for the randomly picked position*/
	short i; /*a loop counter*/
	short j; /*an iteration counter: used to bail out of potentially infinite loops after
	TRIES_PER_COMP times*/

	level[height / 2][width / 2].server = 1; /*place the server in the "center" of the
	level*/

	/*draw a status indicator*/
	FastOutlineRect_R(LCD_MEM, 0, calc(51, 65), calc(159, 239), calc(53, 67), A_NORMAL);
	DrawStr(calc(67, 107), calc(55, 69), "0% done", A_NORMAL);

	for(i = 0; i < computers; i++) /*try to find a spot to place each computer*/
	{
		/*update the progress meter*/
		FastDrawHLine_R(LCD_MEM, 0, calc(159, 239) * (i + 1) / computers, calc(52, 66), A_NORMAL);
		sprintf(buffer, "%d%% done", 100 * (i + 1) / computers);
		DrawStr((calc(160, 240) - DrawStrWidth(buffer, F_4x6)) / 2, calc(55, 69), buffer, A_REPLACE);

		for(j = 0; j < TRIES_PER_COMP; j++) /*try to find a spot TRIES_PER_COMP time for each computer*/
		{
			x = random_long(width);
			y = random_long(height);
			if(SquareIsEmpty(level[y][x])) /*if there is nothing in the square, then it is a valid square
			to place a computer in*/
			{
				break;
			}
		}

		if(j < TRIES_PER_COMP)
		{
			PlaceComputer(x, y, height, width, wrap, level); /*place a computer*/
		}
	}
}

char SquareIsEmpty(SQUARE i) /*returns true if there is nothing in the square*/
{
	return(!(i.server || i.computer || i.con_n || i.con_e || i.con_s || i.con_w));
}

void PlaceComputer(short x, short y, char height, char width, char wrap, SQUARE level[][calc(19, 29)])
{
	char directions[8]; /*the FindValidDirections routine is biased towards the middle of the board, so
	the	directions array requires two entries for each direction.*/
	char old_x = x, old_y = y; /*backups for the start location*/
	char connectable[calc(11, 14)][calc(19, 29)]; /*used to tell if it is possible to connect to a certain
	square*/
	short i; /*keeps track of the number of possibilities that FindValidDirections finds*/
	short j = 0; /*an iteration counter: used to bail out of potentially infinite loops after
	TRIES_PER_COMP times*/
	SQUARE backup[calc(11, 14)][calc(19, 29)]; /*used to store a backup of the level in case placing a
	computer fails*/

	memcpy(backup, level, calc(11, 14) * calc(19, 29) * sizeof(SQUARE)); /*copy the level to a backup so
	it can be restored if placing a computer fails*/
	UpdateConnectable(height, width, level, connectable);
	level[y][x].computer = 1; /*put the computer at the appropriate location*/

	do
	{
		connectable[y][x] = 0;
		i = FindValidDirections(directions, x, y, height, width, wrap, connectable);
		if(i == 0) /*if there are no valid directions left, restore the backup and the start location and
		try again*/
		{
			memcpy(level, backup, calc(11, 14) * calc(19, 29) * sizeof(SQUARE));
			/*restore the level*/

			/*restore the old x and y*/
			x = old_x;
			y = old_y;
			level[(short)y][(short)x].computer = 1; /*put the computer at the appropriate location*/
			UpdateConnectable(height, width, level, connectable);
			j++;
			continue;
		}
	} while(Connect(&x, &y, height, width, directions[random_long(i)], level) && j < TRIES_PER_COMP);
	/*Connect returns false if it has gotten to an appropriate stopping point*/

	if(j >= TRIES_PER_COMP)
	{
		memcpy(level, backup, calc(11, 14) * calc(19, 29) * sizeof(SQUARE)); /*restore the level*/
	}
}

void UpdateConnectable(char height, char width, SQUARE level[][calc(19, 29)],
char connectable[][calc(19, 29)])
{
	short x;
	short y;

	for(x = 0; x < width; x++)
	{
		for(y = 0; y < height; y++)
		{
			if(level[y][x].computer || Connections(level[y][x]) >= 3)
			{
				connectable[y][x] = 0;
			}
			else
			{
				connectable[y][x] = 1;
			}
		}
	}
}

char Connections(SQUARE square)
{
	return(square.con_n + square.con_e + square.con_s + square.con_w);
}

char FindValidDirections(char directions[], short x, short y, char height, char width,
char wrap, char connectable[][calc(19, 29)])
{
	short i = 0; /*the number of valid directions found*/

	memset(directions, 0, 8);

	if((wrap || y > 0) && connectable[Wrap(y - 1, 0, height - 1)][x])
	{
		directions[i++] = 'N';
		if(y > (height / 2))
		{
			directions[i++] = 'N';
		}
	}
	if((wrap || x < width - 1) && connectable[y][Wrap(x + 1, 0, width - 1)])
	{
		directions[i++] = 'E';
		if(x < (width / 2))
		{
			directions[i++] = 'E';
		}
	}
	if((wrap || y < height - 1) && connectable[Wrap(y + 1, 0, height - 1)][x])
	{
		directions[i++] = 'S';
		if(y < (height / 2))
		{
			directions[i++] = 'S';
		}
	}
	if((wrap || x > 0) && connectable[y][Wrap(x - 1, 0, width - 1)])
	{
		directions[i++] = 'W';
		if(x > (width / 2))
		{
			directions[i++] = 'W';
		}
	}

	return i;
}

char Connect(short *xp, short *yp, char height, char width, char direction, SQUARE
level[][calc(19, 29)])
{
	switch(direction)
	{
		case 'N':
		level[*yp][*xp].con_n = 1;
		level[*yp = Wrap((*yp) - 1, 0, height - 1)][*xp].con_s = 1;
		break;

		case 'E':
		level[*yp][*xp].con_e = 1;
		level[*yp][*xp = Wrap((*xp) + 1, 0, width - 1)].con_w = 1;
		break;

		case 'S':
		level[*yp][*xp].con_s = 1;
		level[*yp = Wrap((*yp) + 1, 0, height - 1)][*xp].con_n = 1;
		break;

		case 'W':
		level[*yp][*xp].con_w = 1;
		level[*yp][*xp = Wrap((*xp) - 1, 0, width - 1)].con_e = 1;
	}

	return(Connections(level[*yp][*xp]) < 3 && !level[*yp][*xp].server);
}

void RefreshDisplay(char height, char width, SQUARE level[][calc(19, 29)])
{
	short x;
	short y;

	for(x = 0; x < width; x++)
	{
		for(y = 0; y < height; y++)
		{
			DrawSprite(x, y, height, width, level[y][x]);
		}
	}
}

void UpdateConnected(char height, char width, SQUARE level[][calc(19, 29)])
{
	short x;
	short y;
	short front = 0;
	short back = 0;
	LOCATION queue[QUEUE_SIZE];
	LOCATION temp;

	queue[back++] = (LOCATION){width / 2, height / 2}; /*start at the server*/
	if(back >= QUEUE_SIZE)
	{
		back = 0;
	}

	/*set the whole level to un-connected*/
	for(x = 0; x < width; x++)
	{
		for(y = 0; y < height; y++)
		{
			level[y][x].connected = 0;
		}
	}

	do
	{
		/*get the first entry off the stack and set that square to connected*/
		temp = queue[front++];
		if(front >= QUEUE_SIZE)
		{
			front = 0;
		}
		x = temp.x;
		y = temp.y;
		level[y][x].connected = 1;

		/*if the currect square is connected to any of its horizontal or vertical
		neighbors, and they have not been connected yet, add its neighbors to the stack*/
		if(level[y][x].con_n && !level[Wrap(y - 1, 0, height - 1)][x].connected && level[Wrap(y - 1, 0,
		height - 1)][x].con_s)
		{
			queue[back++] = (LOCATION){x, Wrap(y - 1, 0, height - 1)};
			if(back >= QUEUE_SIZE)
			{
				back = 0;
			}
		}
		if(level[y][x].con_e && !level[y][Wrap(x + 1, 0, width - 1)].connected && level[y][Wrap(x + 1, 0,
		width - 1)].con_w)
		{
			queue[back++] = (LOCATION){Wrap(x + 1, 0, width - 1), y};
			if(back >= QUEUE_SIZE)
			{
				back = 0;
			}
		}
		if(level[y][x].con_s && !level[Wrap(y + 1, 0, height - 1)][x].connected && level[Wrap(y + 1, 0,
		height - 1)][x].con_n)
		{
			queue[back++] = (LOCATION){x, Wrap(y + 1, 0, height - 1)};
			if(back >= QUEUE_SIZE)
			{
				back = 0;
			}
		}
		if(level[y][x].con_w && !level[y][Wrap(x - 1, 0, width - 1)].connected && level[y][Wrap(x - 1, 0,
		width - 1)].con_e)
		{
			queue[back++] = (LOCATION){Wrap(x - 1, 0, width - 1), y};
			if(back >= QUEUE_SIZE)
			{
				back = 0;
			}
		}
	} while(front != back); /*loop until the stack is empty*/
}

char CheckForWin(char height, char width, SQUARE level[][calc(19, 29)])
{
	char win = 1;
	short x;
	short y;

	for(x = 0; x < width; x++)
	{
		for(y = 0; y < height; y++)
		{
			/*all squares must either be empty or connected to win*/
			win &= SquareIsEmpty(level[y][x]) || level[y][x].connected;

			if(!win)
			{
				break;
			}
		}

		if(!win)
		{
			break;
		}
	}

	return(win);
}

short ScrambleLevel(char height, char width, SQUARE level[][calc(19, 29)])
{
	/*just randomly rotates each square that has connections*/
	short rotations;
	short score = 0;
	short x;
	short y;

	for(x = 0; x < width; x++)
	{
		for(y = 0; y < height; y++)
		{
			if(level[y][x].con_n || level[y][x].con_e || level[y][x].con_s || level[y][x].con_w)
			{
				/*if the square is an EW or NS, do not rotate it 180*/
				rotations = random_long((level[y][x].con_n && level[y][x].con_s) || (level[y][x].con_e &&
				level[y][x].con_w) ? 3 : 4) - 1;
				RotateSquare(&level[y][x], rotations);
				score += abs(rotations);
			}
		}
	}

	return(score);
}

void RotateSquare(SQUARE *square, char rotations)
{
	char temp;

	switch(rotations)
	{
		case -1:
		temp = square->con_w;
		square->con_w = square->con_n;
		square->con_n = square->con_e;
		square->con_e = square->con_s;
		square->con_s = temp;
		break;

		case 1:
		temp = square->con_w;
		square->con_w = square->con_s;
		square->con_s = square->con_e;
		square->con_e = square->con_n;
		square->con_n = temp;
		break;

		case 2:
		temp = square->con_n;
		square->con_n = square->con_s;
		square->con_s = temp;
		temp = square->con_w;
		square->con_w = square->con_e;
		square->con_e = temp;
	}
}

void UnscrambleLevel(short *scorep, short *file_scorep, short *computer_movesp, char height, char width,
SQUARE level[][calc(19, 29)])
{
	char buffer[30]; /*a buffer for sprintf*/
	char win = 0; /*wether you have won the game*/
	short key; /*a varibale used to store rowread values*/
	short x = 0;
	short y = 0;

	UpdateConnected(height, width, level);

	level[y][x].cursor = 1;
	RefreshDisplay(height, width, level);

	sprintf(buffer, "  Score: %d Ahead: %d  ", *scorep, *computer_movesp);
	DrawStr((calc(160, 240) - DrawStrWidth(buffer, F_4x6)) / 2, 0, buffer, A_NORMAL);

	do
	{
		key = _rowread(0xFFFE);

		if(key & calc(15, 240)) /*if an arrow key was pressed*/
		{
			/*turn off the cursor*/
			level[y][x].cursor = 0;
			DrawSprite(x, y, height, width, level[y][x]);
		}
		if(key & calc(1, 32)) /*up*/
		{
			y = Wrap(y - 1, 0, height - 1);
			while(_rowread(0xFFFE) & calc(1, 32));
		}
		if(key & calc(8, 64)) /*right*/
		{
			x = Wrap(x + 1, 0, width - 1);
			while(_rowread(0xFFFE) & calc(8, 64));
		}
		if(key & calc(4, 128)) /*down*/
		{
			y = Wrap(y + 1, 0, height - 1);
			while(_rowread(0xFFFE) & calc(4, 128));
		}
		if(key & calc(2, 16)) /*left*/
		{
			x = Wrap(x - 1, 0, width - 1);
			while(_rowread(0xFFFE) & calc(2, 16));
		}
		if(key & calc(15, 240)) /*if an arrow key was pressed*/
		{
			/*turn on the cursor*/
			level[y][x].cursor = 1;
			DrawSprite(x, y, height, width, level[y][x]);
		}

		if(key & calc(16, 1)) /*2nd*/
		{
			RotateSquare(&level[y][x], -1);
			UpdateConnected(height, width, level);
			win = CheckForWin(height, width, level);
			if(win)
			{
				level[y][x].cursor = 0;
			}
			RefreshDisplay(height, width, level);
			*scorep = max((*scorep) - !SquareIsEmpty(level[y][x]), 0);
			(*computer_movesp) -= !SquareIsEmpty(level[y][x]);
			sprintf(buffer, "  Score: %d Ahead: %d  ", *scorep, *computer_movesp);
			DrawStr((calc(160, 240) - DrawStrWidth(buffer, F_4x6)) / 2, 0, buffer, A_REPLACE);
			while(_rowread(0xFFFE) & calc(16, 1));
		}
		if(key & calc(32, 4)) /*shift*/
		{
			RotateSquare(&level[y][x], 1);
			UpdateConnected(height, width, level);
			win = CheckForWin(height, width, level);
			if(win)
			{
				level[y][x].cursor = 0;
			}
			RefreshDisplay(height, width, level);
			*scorep = max((*scorep) - !SquareIsEmpty(level[y][x]), 0);
			(*computer_movesp) -= !SquareIsEmpty(level[y][x]);
			sprintf(buffer, "  Score: %d Ahead: %d  ", *scorep, *computer_movesp);
			DrawStr((calc(160, 240) - DrawStrWidth(buffer, F_4x6)) / 2, 0, buffer, A_REPLACE);
			while(_rowread(0xFFFE) & calc(32, 4));
		}

		if(_rowread(calc(0xFFFD, 0xFF7F)) & calc(64, 32)) /*clear*/
		{
			level[y][x].cursor = 0;
			*file_scorep = *scorep;
			*scorep = -2;
			while(_rowread(calc(0xFFFD, 0xFF7F)) & calc(64, 32));
		}
	} while(!((_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64)) /*esc*/ || win || *scorep == -2));
	while(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64)); /*wait until esc not pressed*/

	if(win)
	{
		FontSetSys(F_8x10);
		FastFilledRect_Erase_R(LCD_MEM, calc(48, 88), calc(44, 58), calc(110, 150), calc(54, 68));
		DrawStr(calc(48, 88), calc(45, 59), "YOU WON!", A_NORMAL);
		FontSetSys(F_4x6);
		Pause();
	}
	else if(*scorep != -2) /*if save game was not pressed, esc must have been pressed*/
	{
		*scorep = -1;
	}
}

void CheckHighScores(short difficulty, short score, short (*HighScores)[3])
{
	short i;
	short j;
	char buffer[10];
	static const char *ordinal[7] = {"1ST", "2ND", "3RD", "4TH", "5TH", "6TH", "7TH"}; /*a
	list of the ordinal numbers*/

	for(i = 0; i < 7; i++)
	{
		if(score > HighScores[i][difficulty])
		{
			FontSetSys(F_8x10);
			FastFilledRect_Erase_R(LCD_MEM, calc(44, 124), calc(39, 53), calc(116, 196), calc(59, 73));
			DrawStr(calc(52, 92), calc(40, 54), "YOU GOT", A_NORMAL);
			sprintf(buffer, "%s PLACE", ordinal[i]);
			DrawStr((calc(160, 240) - 8 * strlen(buffer)) / 2, calc(50, 64), buffer, A_NORMAL);
			FontSetSys(F_4x6);

			for(j = 6; j > i; j--) /*shift the scores up*/
			{
				HighScores[j][difficulty] = HighScores[j - 1][difficulty];
			}
			HighScores[i][difficulty] = score;
			Pause();
			break;
		}
	}
}

void DrawSprite(char x, char y, char height, char width, SQUARE square)
{
	#include "sprites.h" /*the graphics data*/
	short xx = calc(80, 120) - 4 * width + 8 * x, yy = calc(54, 70) - 4 * height + 8 * y;
	/*calculations for centering the sprites horizontally and putting them at the bottom
	of the screen*/

	Sprite8_AND_R(xx, yy, 8, blank, LCD_MEM); /*erase the square*/

	/*the rest is just a ton of ifs to select the correct sprite*/
	if(square.server)
	{
		if(square.con_n)
		{
			if(square.con_e)
			{
				if(square.con_s)
				{
					Sprite8_OR_R(xx, yy, 8, server_nes, LCD_MEM);
				}
				else if(square.con_w)
				{
					Sprite8_OR_R(xx, yy, 8, server_new, LCD_MEM);
				}
				else
				{
					Sprite8_OR_R(xx, yy, 8, server_ne, LCD_MEM);
				}
			}
			else if(square.con_s)
			{
				if(square.con_w)
				{
					Sprite8_OR_R(xx, yy, 8, server_nsw, LCD_MEM);
				}
				else
				{
					Sprite8_OR_R(xx, yy, 8, server_ns, LCD_MEM);
				}
			}
			else if(square.con_w)
			{
				Sprite8_OR_R(xx, yy, 8, server_nw, LCD_MEM);
			}
			else
			{
				Sprite8_OR_R(xx, yy, 8, server_n, LCD_MEM);
			}
		}
		else if(square.con_e)
		{
			if(square.con_s)
			{
				if(square.con_w)
				{
					Sprite8_OR_R(xx, yy, 8, server_esw, LCD_MEM);
				}
				else
				{
					Sprite8_OR_R(xx, yy, 8, server_es, LCD_MEM);
				}
			}
			else if(square.con_w)
			{
				Sprite8_OR_R(xx, yy, 8, server_ew, LCD_MEM);
			}
			else
			{
				Sprite8_OR_R(xx, yy, 8, server_e, LCD_MEM);
			}
		}
		else if(square.con_s)
		{
			if(square.con_w)
			{
				Sprite8_OR_R(xx, yy, 8, server_sw, LCD_MEM);
			}
			else
			{
				Sprite8_OR_R(xx, yy, 8, server_s, LCD_MEM);
			}
		}
		else
		{
			Sprite8_OR_R(xx, yy, 8, server_w, LCD_MEM);
		}
	}
	else if(square.computer)
	{
		if(square.connected)
		{
			if(square.con_n)
			{
				Sprite8_OR_R(xx, yy, 8, computer_n_on, LCD_MEM);
			}
			else if(square.con_e)
			{
				Sprite8_OR_R(xx, yy, 8, computer_e_on, LCD_MEM);
			}
			else if(square.con_s)
			{
				Sprite8_OR_R(xx, yy, 8, computer_s_on, LCD_MEM);
			}
			else if(square.con_w)
			{
				Sprite8_OR_R(xx, yy, 8, computer_w_on, LCD_MEM);
			}
		}
		else
		{
			if(square.con_n)
			{
				Sprite8_OR_R(xx, yy, 8, computer_n, LCD_MEM);
			}
			else if(square.con_e)
			{
				Sprite8_OR_R(xx, yy, 8, computer_e, LCD_MEM);
			}
			else if(square.con_s)
			{
				Sprite8_OR_R(xx, yy, 8, computer_s, LCD_MEM);
			}
			else if(square.con_w)
			{
				Sprite8_OR_R(xx, yy, 8, computer_w, LCD_MEM);
			}
		}
	}
	else if(square.connected)
	{
		if(square.con_n)
		{
			if(square.con_e)
			{
				if(square.con_s)
				{
					Sprite8_OR_R(xx, yy, 8, link_nes_on, LCD_MEM);
				}
				else if(square.con_w)
				{
					Sprite8_OR_R(xx, yy, 8, link_new_on, LCD_MEM);
				}
				else
				{
					Sprite8_OR_R(xx, yy, 8, link_ne_on, LCD_MEM);
				}
			}
			else if(square.con_s)
			{
				if(square.con_w)
				{
					Sprite8_OR_R(xx, yy, 8, link_nsw_on, LCD_MEM);
				}
				else
				{
					Sprite8_OR_R(xx, yy, 8, link_ns_on, LCD_MEM);
				}
			}
			else
			{
				Sprite8_OR_R(xx, yy, 8, link_nw_on, LCD_MEM);
			}
		}
		else if(square.con_e)
		{
			if(square.con_s)
			{
				if(square.con_w)
				{
					Sprite8_OR_R(xx, yy, 8, link_esw_on, LCD_MEM);
				}
				else
				{
					Sprite8_OR_R(xx, yy, 8, link_es_on, LCD_MEM);
				}
			}
			else
			{
				Sprite8_OR_R(xx, yy, 8, link_ew_on, LCD_MEM);
			}
		}
		else
		{
			Sprite8_OR_R(xx, yy, 8, link_sw_on, LCD_MEM);
		}
	}
	else
	{
		if(square.con_n)
		{
			if(square.con_e)
			{
				if(square.con_s)
				{
					Sprite8_OR_R(xx, yy, 8, link_nes, LCD_MEM);
				}
				else if(square.con_w)
				{
					Sprite8_OR_R(xx, yy, 8, link_new, LCD_MEM);
				}
				else
				{
					Sprite8_OR_R(xx, yy, 8, link_ne, LCD_MEM);
				}
			}
			else if(square.con_s)
			{
				if(square.con_w)
				{
					Sprite8_OR_R(xx, yy, 8, link_nsw, LCD_MEM);
				}
				else
				{
					Sprite8_OR_R(xx, yy, 8, link_ns, LCD_MEM);
				}
			}
			else
			{
				Sprite8_OR_R(xx, yy, 8, link_nw, LCD_MEM);
			}
		}
		else if(square.con_e)
		{
			if(square.con_s)
			{
				if(square.con_w)
				{
					Sprite8_OR_R(xx, yy, 8, link_esw, LCD_MEM);
				}
				else
				{
					Sprite8_OR_R(xx, yy, 8, link_es, LCD_MEM);
				}
			}
			else
			{
				Sprite8_OR_R(xx, yy, 8, link_ew, LCD_MEM);
			}
		}
		else if(square.con_s)
		{
			Sprite8_OR_R(xx, yy, 8, link_sw, LCD_MEM);
		}
	}
	if(square.cursor)
	{
		Sprite8_XOR_R(xx, yy, 8, cursor, LCD_MEM);
	}
}

void Pause(void)
{
	while(_rowread(0xFFFE) & calc(16, 1)); /*wait for 2nd to be let up*/
	while(!(_rowread(0xFFFE) & calc(16, 1))); /*wait for 2nd to be pressed*/
	while(_rowread(0xFFFE) & calc(16, 1)); /*wait for 2nd to be let up*/
}
